/* sunmap.c  Turns US Naval Observatory Data */
/* sun azimuth data into a postscript diagram. */
/* Lynn Riordan  April 29 - May 4 2023. */
/* Version 1.1 May 4 2023 */
/*  Added command line input of filename. */
/* Version 1.2 May 8, 2023 */
/*  Added support for Zürich's umlaut. */
/* Version 1.3 May 9, 2023 */
/*  Added support for "Moscow Time" note for Kerch Bridge. */
/* Version 1.4 May 15, 2023 */
/*  Added typeface option -t */
/* Version 2.0 May 19, 2023 */
/*  Added ability to parse raw html code. */
/*  Like the kind of files cURL returns. */
/* Version 2.1 May 20, 2023 */
/*  Deleted my buggy findtext() and replaced with strstr(). */
/*  Separated html and plain text file processing. */
/*  Created htmlparse() seperate from parse() for this. */
/* Version 2.2 May 29, 2023 */
/*  Added -C color option. */
/* Version 2.3 June 8-12, 2023 */
/*  Added workaround to having spaces in names in cURL requests rejected. */
/* Version 3.0 June 26 2023 */
/*  Changed order of data reading in switch() in htmlparse() to */
/*  match what the Naval Observatory changed it to. */
/*
 *  Copyright (C) 2023 Lynn Riordan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>  /* fgets(), fopen(), fprintf() */
#include <string.h> /* strlen(), strcpy(), strstr() */
#include <stdlib.h> /* atoi() */
#include <ctype.h>  /* isdigit() */
#include <math.h> /* double sin(double), cos(double) */

/* #define DOS */
#define MAJORVER 3   /* So I don't have to run all over the code */
#define MINORVER 0   /* everytime I make a new version. */
#define TRUE 1
#define FALSE 0
#define BUFFERSIZE 256
#define CR 13
#define LF 10
#define SP 32
#define TAB   9
#define DEGREE 176
#define DEGTORAD(x) ((x) / 57.29578)
#define RADIUS 252  /* Length of radius in 72nds of an inch. */
#define CTRX 286    /* x coordinate of center of circle in 72nds of an inch. */
#define CTRY 324    /* y coordinate of center of circle in 72nds of an inch. */
#define SLICES 96   /* 15 minutes out of a 24 hour clock. */

char BIGHEADER[BUFFERSIZE];
char LITTLEHEADER[BUFFERSIZE];
char typeface[BUFFERSIZE];
int CMDLINE = FALSE;
int ZURICH = FALSE;
int KERCH = FALSE;
int TYPEFACE = FALSE;
int HTML = FALSE;
int MOON = FALSE;
int COLOR = FALSE;
int DEBUG = FALSE;

void help(char * programName);  /* Tells how to use this utility. */
unsigned long parse(FILE * fpIn, FILE * fpOut); /* Work done here. */
unsigned long htmlparse(FILE * fpIn, FILE * fpOut); /* Work done here. */
void textextract(char * dest, char * src); /* Gets text out of html. Does modify src. */


int main(int argc, char * argv[])
  {
  FILE * fpIn;
  FILE * fpOut;    /*  are written here. */
  unsigned long total = 0;
  int i, j;
  char buffer[BUFFERSIZE];
  char programName[BUFFERSIZE];

  /* The following for() loop gives help() just this program's name, without the path. */
  for( i = 0, j = 0; argv[0][i] != 0; i++)
    {
    buffer[j++] = argv[0][i];
#ifdef DOS
    if(argv[0][i] == '\\')
#else
    if(argv[0][i] == '/')
#endif
    j = 0;
    }
  buffer[j] = 0;
#ifdef DOS
  for( i = 0; i < BUFFERSIZE; i++)
    {
    if(buffer[i] == '.')
      {
      buffer[i] = 0;
      break;
      }
    }
#endif

  for(i = 0; i < strlen(buffer); i++)
    programName[i] = buffer[i];

  programName[i] = 0;

  /* Now we check for command line options. */
  for( i = 1, j = 0; i < argc; i++)
    {
    if( (argv[i][0] == '/') || (argv[i][0] == '-') )
      {
      switch(argv[i][1])
        {
        case '?' :
          help(programName);
          return 0;
        case 'C' :
          printf("\nOption C, color option invoked!\n");
          COLOR = TRUE;
          break;
        case 'D' :
          printf("\nOption Debug invoked!\n");
          DEBUG = TRUE;
          break;
        case 'F' :
          if((fpIn = fopen((char *) &(argv[i][2]), "r")) == NULL)
            {
            fprintf(stderr, "ERROR: Can't open input file %s\n", (char *) &(argv[i][2]));
            return 1;
            }
          else
            {
            CMDLINE = TRUE;
            printf("\n%s opened for input.\n\n", (char *) &(argv[i][2]));
            strncpy(buffer, (char *) &(argv[i][2]), BUFFERSIZE);
            }
          break;
        case 'H' :
          help(programName);
          return 0;
        case 'T' :
          printf("Typeface option invoked!\n");
          TYPEFACE = TRUE;
          strncpy(typeface,  (char *) &(argv[i][2]), BUFFERSIZE);
          break;
        case 'W':
          printf("HTML option invoked!\n");
          HTML = TRUE;
          break;
        case 'Z':
          printf("Zürich option invoked!\n");
          ZURICH = TRUE;
          break;
        case 'c' :
          printf("\nOption c, color option, invoked!\n");
          COLOR = TRUE;
          break;
        case 'd' :
          printf("Option d (debug) invoked!\n");
          DEBUG = TRUE;
          break;
        case 'f' :
          if((fpIn = fopen((char *) &(argv[i][2]), "r")) == NULL)
            {
            fprintf(stderr, "ERROR: Can't open input file %s\n", (char *) &(argv[i][2]));
            return 1;
            }
          else
            {
            CMDLINE = TRUE;
            printf("\n%s opened for input.\n\n", (char *) &(argv[i][2]));
            strncpy(buffer, (char *) &(argv[i][2]), BUFFERSIZE);
            }
          break;
        case 'h' :
          help(programName);
          return 0;
        case 't' :
          printf("Typeface option invoked!\n");
          TYPEFACE = TRUE;
          strncpy(typeface,  (char *) &(argv[i][2]), BUFFERSIZE);
          break;
        case 'w':
          printf("HTML option invoked!\n");
          HTML = TRUE;
          break;
        case 'z':
          printf("Zürich option invoked!\n");
          ZURICH = TRUE;
          break;
        default:
          fprintf(stderr, "ERROR: Unknown command line option %s invoked. Ignoring.\n", argv[i]);
        } /* end switch(argv[i][1]) */
      } /* end if() */
    } /* end for(). End command line option checking. */

  printf("\n%s - Sun position diagram utility, version %d.%d\n", programName, MAJORVER, MINORVER);
  printf("Copyright 2023 Lynn Riordan under GNU General Public License version 3 or later.\n\n");

  if(CMDLINE == FALSE)
    {
    printf("Enter name of input file:\n");
    fgets(buffer, BUFFERSIZE, stdin);
    for( i = 0; i < BUFFERSIZE; i++)
      {
      if( (buffer[i] == LF) || (buffer[i] == CR) )
        {
        buffer[i] = 0;
        break;
        }
      } /* End for( i = 0; i < BUFFERSIZE; i++) */

    if((fpIn = fopen(buffer, "r")) == NULL)
      {
      fprintf(stderr, "ERROR: Can't open input file %s\n", buffer);
      return 1;
      }
    else
      printf("\n%s opened for input.\n\n", buffer);
    } /* End if(CMDLINE == FALSE) */

  for( i = 0; i < BUFFERSIZE; i++)
    {
    if((buffer[i] == '.') || (buffer[i] == 0))
      {
      break;
      }
    }
  if(i > (BUFFERSIZE - 5))
    strncpy((char *) &(buffer[BUFFERSIZE - 5]), ".ps", 3);
  else
    strncpy((char *) &(buffer[i]), ".ps", 3);
    
  i += 3;
  buffer[i] = 0;

  if((fpOut = fopen(buffer, "w")) == NULL)
    {
    fprintf(stderr, "ERROR: Can't open output file %s\n", buffer);
    return 1;
    }
  else
    printf("%s opened for output in overwrite mode.\n", buffer);


  total = parse(fpIn, fpOut);
  fprintf(fpOut, "showpage\n");
  fclose(fpIn);
  fclose(fpOut);
  printf("\nTotal number of records converted: %lu\n\n", total);

  return 0;
  }

void help(char * programName)
  {
  printf("\n%s - Utility for making diagrams out of sun data. Version %d.%d\n", programName, MAJORVER, MINORVER);
  printf("Copyright 2023 Lynn Riordan under GNU General Public License version 3 or later.\n");
  printf("\tUSAGE: %s [-H][-F<inputfile>][-T<typeface>][-C][-W][-Z]\n", programName);
  printf("\t-H option invokes help() and then quits.\n");
  printf("\t-F<inputfilename> allows one to give input filename on command line.\n");
  printf("\t-T<Name of typeface> Changes header font from Optima to typeface.\n");
  printf("\t-C makes the program output purple Postscript.\n");
  printf("\t-W forces the program to treat the input as HTML.\n");
  printf("\t-Z option forces the program to treat the input as being about Zürich.\n");
  printf("\nThis program takes the given input file and creates\n");
  printf("an output file with a .ps file extenstion.\n");
  printf("\nThis program is free software; you can redistribute it and/or modify\n");
  printf("it under the terms of the GNU General Public License as published by\n");
  printf("the Free Software Foundation; either version 3 of the License, or\n");
  printf("(at your option) any later version.\n\n");

  printf("This program is distributed in the hope that it will be useful,\n");
  printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
  printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
  printf("GNU Library General Public License, in Copying.txt, for more details.\n\n");
 
  printf("You should have received a copy of the GNU General Public License\n");
  printf("(Copying.txt) along with this program; if not, write to the Free Software\n");
  printf("Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.\n");
  return;
  }


unsigned long parse(FILE * fpIn, FILE * fpOut)
  {
  int i = 0;      /* Loop control variables. */
  int j;
  int k;
  int state = 0;  /* State variable for finite state machine. */
  unsigned long total = 0;  /* Total number of records read in. */
  char buffer[BUFFERSIZE];
  char htmlbuffer[BUFFERSIZE];
  char time[6];   /* Holds the time label. */
  double altitude;
  double azimuth; /* Holds angle value of radius being drawn. */
  double x,y;     /* Coordinates of ending point of line drawn. */
  double x2,y2;   /* Starting points for short lines. */
  double smallfont = 9.0;  /* Size of labels. */
  double temp;

  /* Write header to PostScript file. */
  fprintf( fpOut, "%%!PS\n");
  fprintf( fpOut, "%%%%BoundingBox: 0 20 612 812\n");
  fprintf( fpOut, "%%Postscript file created by sunmap.c, Version %d.%d by Lynn Riordan.\n", MAJORVER, MINORVER);
  fprintf( fpOut, "%%Copyright 2023 Lynn Riordan under GNU General Public License version 3 or later.\n");
  if(TYPEFACE)
    fprintf( fpOut, "/%s findfont\n", typeface);
  else
    fprintf( fpOut, "/Optima findfont\n");

  fprintf( fpOut, "64 scalefont\n");
  fprintf( fpOut, "setfont\n");
  fprintf( fpOut, "newpath\n");
  if(COLOR)
    fprintf( fpOut, "0.5 0.0 0.5 setrgbcolor\n");
  fprintf( fpOut, "/cshow { dup stringwidth pop -0.5\n");
  fprintf( fpOut, "mul 0 rmoveto show } def\n");
  fprintf( fpOut, "306 747.68 moveto\n\n\n");
 

  while( fgets(buffer, BUFFERSIZE, fpIn) != NULL)
    {
    if(strncmp(buffer, "<!doctype html>", 15) == 0)
      {
      HTML = TRUE;
      printf("HTML header found.\n");
      total = htmlparse( fpIn, fpOut);
      return total;
      }
    switch(state)
      {
      case 0:
        if(strncmp(buffer, "Altitude", 8) == 0)
            {
            if(strstr(buffer, "Moon") != NULL)
              {
              MOON = TRUE;
              printf("\nMoon data!\n");
              }
            state++;
            }
        break;
      case 1:
        for( i = 0; i < BUFFERSIZE; i++)
          {
          if( (buffer[i] == LF) || (buffer[i] == CR) )
            {
            buffer[i] = 0;
            if(buffer[--i] == SP)
              buffer[i] = 0;
            break;
            }
          }

        if(strstr(buffer, "ZURICH") != NULL)
          ZURICH = TRUE;
        if(strstr(buffer, "Zurich") != NULL)
          ZURICH = TRUE;
        if(strstr(buffer, "zurich") != NULL)
          ZURICH = TRUE;

        if(strstr(buffer, "KERCH") != NULL)
          KERCH = TRUE;
        if((strstr(buffer, "Kerch") != NULL) || (strstr(buffer, "kerch") != NULL) )
          KERCH = TRUE;

        if(ZURICH)
          {
          fprintf( fpOut, "217 736.88 moveto\n");
          fprintf( fpOut, "(Z) show /udieresis glyphshow\n");
          fprintf( fpOut, "(rich) show\n");
          fprintf( fpOut, "newpath\n");
          fprintf( fpOut, "2.0 setlinewidth\n");
          fprintf( fpOut, "306 736.88 moveto\n");
          strcpy(BIGHEADER, "Zurich");
          }
        else
          {
          strcpy(BIGHEADER, buffer);
    if(strcmp("CapeTown", BIGHEADER) == 0)
      strcpy(BIGHEADER, "Cape Town");
    if(strcmp("Kerch", BIGHEADER) == 0)
      strcpy(BIGHEADER, "Kerch Bridge");
    if(strcmp("LA", BIGHEADER) == 0)
      strcpy(BIGHEADER, "Los Angeles");
    if(strcmp("McMurdo", BIGHEADER) == 0)
      strcpy(BIGHEADER, "McMurdo Station");
    if(strcmp("NYC", BIGHEADER) == 0)
      strcpy(BIGHEADER, "New York City");
    if(strcmp("SF", BIGHEADER) == 0)
      strcpy(BIGHEADER, "San Francisco");
          fprintf( fpOut, "(%s) cshow\n", BIGHEADER);
          fprintf( fpOut, "newpath\n");
          fprintf( fpOut, "2.0 setlinewidth\n");
          fprintf( fpOut, "306 736.88 moveto\n");
          }

        fprintf( fpOut, "(%s) dup stringwidth pop -0.5\n", BIGHEADER);
        fprintf( fpOut, "mul 0 rmoveto\n");
        fprintf( fpOut, "(%s) dup stringwidth pop\n", BIGHEADER);
        fprintf( fpOut, "0 rlineto\n");
        fprintf( fpOut, "stroke\n\n");

        if(TYPEFACE)
          fprintf( fpOut, "/%s findfont\n", typeface);
        else
          fprintf( fpOut, "/Optima findfont\n");

        fprintf( fpOut, "48 scalefont\n");
        fprintf( fpOut, "setfont\n");
        fprintf( fpOut, "newpath\n");
        fprintf( fpOut, "306 685.04 moveto\n\n");
        state++;
        break;
      case 2:
        if(strncmp(buffer, "202", 3) == 0)
          {
          for( i = 0; i < BUFFERSIZE; i++)
            {
            if( (buffer[i] == LF) || (buffer[i] == CR) )
              {
              buffer[i] = 0;
              if(buffer[--i] == SP)
                buffer[i] = 0;
               break;
               } /* End if( (buffer[i] == LF) || (buffer[i] == CR) ) */
             } /* End for( i = 0; i < BUFFERSIZE; i++) */
           for( i = 0; i < 10; i++)
             {
             LITTLEHEADER[i] = buffer[i];
             } /* End for() */
            LITTLEHEADER[i] = 0;

          /* Write rest of header. */
          fprintf( fpOut, "(%s) cshow\n", LITTLEHEADER);
          if(KERCH)
            {
            if(TYPEFACE)
              fprintf( fpOut, "/%s findfont\n", typeface);
            else
              fprintf( fpOut, "/Optima findfont\n");
            fprintf( fpOut, "10 scalefont\n");
            fprintf( fpOut, "setfont\nnewpath\n");
            fprintf( fpOut, "306 635 moveto\n");
            fprintf( fpOut, "(Moscow timezone) show\n");
            if(TYPEFACE)
              fprintf( fpOut, "/%s findfont\n", typeface);
            else
              fprintf( fpOut, "/Optima findfont\n");
            fprintf( fpOut, "48 scalefont\n");
            fprintf( fpOut, "setfont\nnewpath\n");
            } /* End if(KERCH) */

          fprintf( fpOut, "%d %d moveto\n", CTRX - 18, CTRY + RADIUS);
          fprintf( fpOut, "(N) show\n");
          fprintf( fpOut, "newpath\n");
          fprintf( fpOut, "1.5 setlinewidth\n");
          fprintf( fpOut, "306 672.08 moveto\n");
          fprintf( fpOut, "(%s) dup stringwidth pop -0.5\n", LITTLEHEADER);
          fprintf( fpOut, "mul 0 rmoveto\n");
          fprintf( fpOut, "(%s) dup stringwidth pop\n", LITTLEHEADER);
          fprintf( fpOut, "0 rlineto\n");
          fprintf( fpOut, "stroke\n\n");
          fprintf( fpOut, "/Times-Roman findfont\n");
          fprintf( fpOut, "%.1f scalefont\n", smallfont);
          fprintf( fpOut, "setfont\n");
          fprintf( fpOut, "%%Translate coordinate system to one where (0,0)\n");
          fprintf( fpOut, "%%is the center of the circle. It'll remain in 72'nds of an inch.\n");
          fprintf( fpOut, "%d %d translate\n\n", CTRX, CTRY);
          fprintf( fpOut, "%%Draw circle.\n");
          fprintf( fpOut, "newpath\n");
          fprintf( fpOut, "0 0 %d 0 360 arc\n", RADIUS);
          fprintf( fpOut, "3.0 setlinewidth\n");
          fprintf( fpOut, "0.3 setgray\n");
          if(COLOR)
            fprintf( fpOut, "0.5 0.0 0.5 setrgbcolor\n");
          fprintf( fpOut, "stroke\n\n");

          fprintf( fpOut, "0.0 setgray\n");
          if(COLOR)
            fprintf( fpOut, "0.5 0.0 0.5 setrgbcolor\n");
          fprintf( fpOut, "2.5 setlinewidth\n");
          fprintf( fpOut, "0 0 moveto\n");
          fprintf( fpOut, "%d %d lineto\n", 0, RADIUS);
          fprintf( fpOut, "stroke\n\n");

          buffer[1] = 0;
          buffer[0] = 0;
          i = 0;
          state++;
          } /* End if(strncmp(buffer, "202", 3) == 0) */
        break;
      case 3:  /* Where we start finding records. */
        if((buffer[0] == '0') || (buffer[0] == '1') || (buffer[0] == '2'))
          {
          strncpy(time, buffer, 5);
          time[5] = 0;
          i = 6;
          altitude = strtod(&(buffer[i]), NULL);
          while(buffer[i] != TAB)
            i++;
          i++;
          azimuth = strtod(&(buffer[i]), NULL);
          total++;
          /* Convert from E of North style to math style. */
          azimuth *= -1.0;
          azimuth += 90.0;
          temp = fmod(azimuth, 360.0);
          azimuth = temp;
          x = cos( DEGTORAD(azimuth)) * RADIUS;
          y = sin( DEGTORAD(azimuth)) * RADIUS;
          if( ((time[3] == '1') || (time[3] == '4')) && (time[4] == '5') )
            {
            fprintf( fpOut, "0.5 setlinewidth\n");
            fprintf( fpOut, "newpath\n");
            x2 = cos( DEGTORAD(azimuth)) * RADIUS * 0.95;
            y2 = sin( DEGTORAD(azimuth)) * RADIUS * 0.95;
            fprintf( fpOut, "%.1f %.1f moveto\n", x2, y2);
            fprintf( fpOut, "%.1f %.1f lineto\n", x, y);
            fprintf( fpOut, "stroke\n");
            }
          else if((time[3] == '0') && (time[4] == '0'))
            {
            fprintf( fpOut, "1.5 setlinewidth\n");
            fprintf( fpOut, "newpath\n");
            fprintf( fpOut, "0 0 moveto\n");
            fprintf( fpOut, "%.1f %.1f lineto\n", x, y);
            fprintf( fpOut, "stroke\n");
            }
          else if((time[3] == '3') && (time[4] == '0'))
            {
            fprintf( fpOut, "0.75 setlinewidth\n");
            fprintf( fpOut, "newpath\n");
            fprintf( fpOut, "0 0 moveto\n");
            fprintf( fpOut, "%.1f %.1f lineto\n", x, y);
            fprintf( fpOut, "stroke\n");
            }
          else
            {
            fprintf(fpOut, "0.25 setlinewidth\n");
            fprintf( fpOut, "newpath\n");
            x2 = cos( DEGTORAD(azimuth)) * RADIUS * 0.97;
            y2 = sin( DEGTORAD(azimuth)) * RADIUS * 0.97;
            fprintf( fpOut, "%.1f %.1f moveto\n", x2, y2);
            fprintf( fpOut, "%.1f %.1f lineto\n", x, y);
            fprintf( fpOut, "stroke\n");
            }

          fprintf(fpOut, "gsave\n\n");
          temp = (time[0] - '0') * 10.0;
          temp += (time[1] - '0');
          if(((time[3] == '3') && (time[4] == '0')) || ((time[3] == '0') && (time[4] == '0')))
            {
            if((azimuth <= 90.0) && (azimuth >= -90.0))
              { /* Right side */
              x = cos(DEGTORAD(azimuth + (1))) * (RADIUS * 0.91);
              y = sin(DEGTORAD(azimuth + (1))) * (RADIUS * 0.91);
              fprintf(fpOut, "%.1f %.1f translate\n", x, y);
              fprintf(fpOut, "%.1f rotate\n", azimuth + (1));
              fprintf(fpOut, "newpath\n");
              fprintf(fpOut, "0 0 moveto\n");
              fprintf(fpOut, "(%s   %.1f deg.) show\n", time, altitude);
              fprintf(fpOut, "grestore\n");
              }
            else
              { /* Left side */
              x = cos(DEGTORAD(azimuth + (2.0))) * RADIUS * 1.07;
              y = sin(DEGTORAD(azimuth + (2.0))) * RADIUS * 1.07;
              fprintf(fpOut, "%.1f %.1f translate\n", x, y);
              fprintf(fpOut, "%.1f rotate\n", azimuth - 179.0);
              fprintf(fpOut, "newpath\n");
              fprintf(fpOut, "0 0 moveto\n");
              fprintf(fpOut, "(%.1f deg. %s) show\n", altitude, time);
              fprintf(fpOut, "grestore\n");
              }
            }
          }
       break;
      }  /* End switch(state) */
   } /* End while( (ch = getc(fpIn)) != EOF) */
   

  return  total;
  }


void textextract(char * dest, char * src)
  {
  int i, j, state;

  /* Fix src from the back end */
  for(i = 0; i < BUFFERSIZE; i++)
    {
    if( (src[i] == LF) || (src[i] == CR) )
      {
      src[i] = 0;
      if(src[--i] == SP)
        src[i] = 0;
      break;
      } /* End if( (src[i] == LF) || (src[i] == CR) ) */
    } /* End for(i = 0; i < BUFFERSIZE; i++) */

  /* Copy saner text from the beginning */
  for( i = 0, state = 0, j = 0; i < strlen(src); i++)
    {
    switch(state)
      {
      case 0:
        if(src[i] != SP)
          state++;
        break;
      case 1:
        if(src[i] == '>')
          state++;
        break;
      case 2:
        if(src[i] != SP)
          {
          dest[j++] = src[i];
          state++;
          }
        break;
      case 3:
        dest[j++] = src[i];
        if(src[i] == 0)
          return;
        break;
      } /* End switch() */
    } /* End for() */

  return;
  }


unsigned long htmlparse(FILE * fpIn, FILE * fpOut)
  {
  char buffer[BUFFERSIZE];
  char htmlbuffer[BUFFERSIZE];
  int i;
  int state = 0;
  unsigned long total = 0;  /* Total number of records read in. */
  char time[6];   /* Holds the time label. */
  double altitude;
  double azimuth; /* Holds angle value of radius being drawn. */
  double x,y;     /* Coordinates of ending point of line drawn. */
  double x2,y2;   /* Starting points for short lines. */
  double smallfont = 9.0;  /* Size of labels. */
  double temp;


  while( fgets(buffer, BUFFERSIZE, fpIn) != NULL)
    {
    for(i = 0; i < BUFFERSIZE; i++)
      htmlbuffer[i] = 0;
    textextract(htmlbuffer, buffer);
    switch(state)
      {
      case 0:
        if(strstr(buffer, "<th colspan=") != NULL)
          {
          if(DEBUG)
            {
            printf("Case 0. buffer =%s\n", buffer);
            printf("htmlbuffer =%s\n", htmlbuffer);
            }
          if(strstr(buffer, "Moon") != NULL)
            {
            MOON = TRUE;
            printf("\nMoon data!\n");
            }
          state++;
          }
        break;
      case 1: // htmlparse()
        if(strstr(htmlbuffer, "ZURICH") != NULL)
          ZURICH = TRUE;
        if(strstr(htmlbuffer, "Zurich") != NULL)
          ZURICH = TRUE;
        if(strstr(htmlbuffer, "zurich") != NULL)
          ZURICH = TRUE;

        if(strstr(htmlbuffer, "KERCH") != NULL)
          KERCH = TRUE;
        if((strstr(htmlbuffer, "Kerch") != NULL) || (strstr(htmlbuffer, "kerch") != NULL) )
          {
          strcpy(htmlbuffer, "Kerch Bridge");
          KERCH = TRUE;
          }
        if(COLOR)
          fprintf( fpOut, "0.5 0.0 0.5 setrgbcolor\n");

        if(ZURICH)
          {
          fprintf( fpOut, "217 736.88 moveto\n");
          fprintf( fpOut, "(Z) show /udieresis glyphshow\n");
          fprintf( fpOut, "(rich) show\n");
          fprintf( fpOut, "newpath\n");
          fprintf( fpOut, "2.0 setlinewidth\n");
          fprintf( fpOut, "306 736.88 moveto\n");
          strcpy(BIGHEADER, "Zurich");
          }
        else
          {
          strcpy(BIGHEADER, htmlbuffer);
          if(strcmp("CapeTown", BIGHEADER) == 0)
            strcpy(BIGHEADER, "Cape Town");
          if(strcmp("Kerch", BIGHEADER) == 0)
            strcpy(BIGHEADER, "Kerch Bridge");
          if(strcmp("LA", BIGHEADER) == 0)
            strcpy(BIGHEADER, "Los Angeles");
          if(strcmp("McMurdo", BIGHEADER) == 0)
            strcpy(BIGHEADER, "McMurdo Station");
          if(strcmp("NYC", BIGHEADER) == 0)
            strcpy(BIGHEADER, "New York City");
          if(strcmp("SF", BIGHEADER) == 0)
            strcpy(BIGHEADER, "San Francisco");
          fprintf( fpOut, "(%s) cshow\n", BIGHEADER);
          fprintf( fpOut, "newpath\n");
          fprintf( fpOut, "2.0 setlinewidth\n");
          fprintf( fpOut, "306 736.88 moveto\n");
          }
        fprintf( fpOut, "(%s) dup stringwidth pop -0.5\n", BIGHEADER);
        fprintf( fpOut, "mul 0 rmoveto\n");
        fprintf( fpOut, "(%s) dup stringwidth pop\n", BIGHEADER);
        fprintf( fpOut, "0 rlineto\n");
        fprintf( fpOut, "stroke\n\n");

        if(TYPEFACE)
          fprintf( fpOut, "/%s findfont\n", typeface);
        else
          fprintf( fpOut, "/Optima findfont\n");

        fprintf( fpOut, "48 scalefont\n");
        fprintf( fpOut, "setfont\n");
        fprintf( fpOut, "newpath\n");
        fprintf( fpOut, "306 685.04 moveto\n\n");
        state++;

        break;
      case 2: // htmlparse()
        if( strstr(htmlbuffer, "202") != NULL)
          {
          for( i = 0; i < 10; i++)
            {
            LITTLEHEADER[i] = htmlbuffer[i];
            } /* End for() */
          LITTLEHEADER[i] = 0;
          state++;
          /* Write rest of header. */
          fprintf( fpOut, "(%s) cshow\n", LITTLEHEADER);
          if(KERCH)
            {
            if(TYPEFACE)
              fprintf( fpOut, "/%s findfont\n", typeface);
            else
              fprintf( fpOut, "/Optima findfont\n");
            fprintf( fpOut, "10 scalefont\n");
            fprintf( fpOut, "setfont\nnewpath\n");
            fprintf( fpOut, "306 635 moveto\n");
            fprintf( fpOut, "(Moscow timezone) show\n");
            if(TYPEFACE)
              fprintf( fpOut, "/%s findfont\n", typeface);
            else
              fprintf( fpOut, "/Optima findfont\n");
            fprintf( fpOut, "48 scalefont\n");
            fprintf( fpOut, "setfont\nnewpath\n");
            } /* End if(KERCH) */
 // htmlparse()
          fprintf( fpOut, "%d %d moveto\n", CTRX - 18, CTRY + RADIUS);
          fprintf( fpOut, "(N) show\n");
          fprintf( fpOut, "newpath\n");
          fprintf( fpOut, "1.5 setlinewidth\n");
          fprintf( fpOut, "306 672.08 moveto\n");
          fprintf( fpOut, "(%s) dup stringwidth pop -0.5\n", LITTLEHEADER);
          fprintf( fpOut, "mul 0 rmoveto\n");
          fprintf( fpOut, "(%s) dup stringwidth pop\n", LITTLEHEADER);
          fprintf( fpOut, "0 rlineto\n");
          fprintf( fpOut, "stroke\n\n");
          fprintf( fpOut, "/Times-Roman findfont\n");
          fprintf( fpOut, "%.1f scalefont\n", smallfont);
          fprintf( fpOut, "setfont\n");
          fprintf( fpOut, "%%Translate coordinate system to one where (0,0)\n");
          fprintf( fpOut, "%%is the center of the circle. It'll remain in 72'nds of an inch.\n");
          fprintf( fpOut, "%d %d translate\n\n", CTRX, CTRY);
          fprintf( fpOut, "%%Draw circle.\n");
          fprintf( fpOut, "newpath\n");
          fprintf( fpOut, "0 0 %d 0 360 arc\n", RADIUS);
          fprintf( fpOut, "3.0 setlinewidth\n");
          fprintf( fpOut, "0.3 setgray\n");
          if(COLOR)
            fprintf( fpOut, "0.5 0.0 0.5 setrgbcolor\n");
          fprintf( fpOut, "stroke\n\n");
          fprintf( fpOut, "0.0 setgray\n");
          if(COLOR)
            fprintf( fpOut, "0.5 0.0 0.5 setrgbcolor\n");
          fprintf( fpOut, "2.5 setlinewidth\n");
          fprintf( fpOut, "0 0 moveto\n");
          fprintf( fpOut, "%d %d lineto\n", 0, RADIUS);
          fprintf( fpOut, "stroke\n\n");

          buffer[1] = 0;
          buffer[0] = 0;
          i = 0;
          state++;
          } /* End if(strncmp(buffer, "202", 3) == 0) */
        break;
      case 3: // htmlparse()
        if((htmlbuffer[0] == '-') || (isdigit(htmlbuffer[0])) )
          {
          azimuth = strtod(htmlbuffer, NULL);
          /* Convert from E of North style to math style. */
          azimuth *= -1.0;
          azimuth += 90.0;
          temp = fmod(azimuth, 360.0);
          azimuth = temp;
          if(DEBUG)
            {
            printf("Case 3. buffer =%s\n", buffer);
            printf("htmlbuffer =%s\n", htmlbuffer);
            printf("azimuth =%f\n", azimuth);
            }
          state++;
          }
        break;
      case 4:
        if((htmlbuffer[0] == '0') || (htmlbuffer[0] == '1') || (htmlbuffer[0] == '2'))
          {
          for(i =0; i < 5; i++)
            time[i] = htmlbuffer[i];
          time[5] = 0;
          if(DEBUG)
            {
            printf("Case 4. buffer =%s\n", buffer);
            printf("htmlbuffer =%s\n", htmlbuffer);
            printf("time =%s\n", time);
            }
          state++;
          }
        break;
      case 5:
        if( (htmlbuffer[0] == '-') || (isdigit(htmlbuffer[0])) )
          {
          altitude = strtod(htmlbuffer, NULL);
          if(DEBUG)
            {
            printf("Case 5. buffer =%s\n", buffer);
            printf("htmlbuffer =%s\n", htmlbuffer);
            printf("Altitude =%f\n", altitude);
            }
          /* Write the record out. */
          x = cos( DEGTORAD(azimuth)) * RADIUS;
          y = sin( DEGTORAD(azimuth)) * RADIUS;
          if( ((time[3] == '1') || (time[3] == '4')) && (time[4] == '5') )
            {
            fprintf( fpOut, "0.5 setlinewidth\n");
            fprintf( fpOut, "newpath\n");
            x2 = cos( DEGTORAD(azimuth)) * RADIUS * 0.95;
            y2 = sin( DEGTORAD(azimuth)) * RADIUS * 0.95;
            fprintf( fpOut, "%.1f %.1f moveto\n", x2, y2);
            fprintf( fpOut, "%.1f %.1f lineto\n", x, y);
            fprintf( fpOut, "stroke\n");
            }
          else if((time[3] == '0') && (time[4] == '0'))
            {
            fprintf( fpOut, "1.5 setlinewidth\n");
            fprintf( fpOut, "newpath\n");
            fprintf( fpOut, "0 0 moveto\n");
            fprintf( fpOut, "%.1f %.1f lineto\n", x, y);
            fprintf( fpOut, "stroke\n");
            }
          else if((time[3] == '3') && (time[4] == '0'))
            {
            fprintf( fpOut, "0.75 setlinewidth\n");
            fprintf( fpOut, "newpath\n");
            fprintf( fpOut, "0 0 moveto\n");
            fprintf( fpOut, "%.1f %.1f lineto\n", x, y);
            fprintf( fpOut, "stroke\n");
            }
          else
            {
            fprintf(fpOut, "0.25 setlinewidth\n");
            fprintf( fpOut, "newpath\n");
            x2 = cos( DEGTORAD(azimuth)) * RADIUS * 0.97;
            y2 = sin( DEGTORAD(azimuth)) * RADIUS * 0.97;
            fprintf( fpOut, "%.1f %.1f moveto\n", x2, y2);
            fprintf( fpOut, "%.1f %.1f lineto\n", x, y);
            fprintf( fpOut, "stroke\n");
            }

          fprintf(fpOut, "gsave\n\n");
          temp = (time[0] - '0') * 10.0;
          temp += (time[1] - '0');
          if(((time[3] == '3') && (time[4] == '0')) || ((time[3] == '0') && (time[4] == '0')))
            {
            if((azimuth <= 90.0) && (azimuth >= -90.0))
              { /* Right side */
              x = cos(DEGTORAD(azimuth + (1))) * (RADIUS * 0.91);
              y = sin(DEGTORAD(azimuth + (1))) * (RADIUS * 0.91);
              fprintf(fpOut, "%.1f %.1f translate\n", x, y);
              fprintf(fpOut, "%.1f rotate\n", azimuth + (1));
              fprintf(fpOut, "newpath\n");
              fprintf(fpOut, "0 0 moveto\n");
              fprintf(fpOut, "(%s   %.1f deg.) show\n", time, altitude);
              fprintf(fpOut, "grestore\n");
              }
            else
              { /* Left side */
              x = cos(DEGTORAD(azimuth + (2.0))) * RADIUS * 1.07;
              y = sin(DEGTORAD(azimuth + (2.0))) * RADIUS * 1.07;
              fprintf(fpOut, "%.1f %.1f translate\n", x, y);
              fprintf(fpOut, "%.1f rotate\n", azimuth - 179.0);
              fprintf(fpOut, "newpath\n");
              fprintf(fpOut, "0 0 moveto\n");
              fprintf(fpOut, "(%.1f deg. %s) show\n", altitude, time);
              fprintf(fpOut, "grestore\n");
              }
            }

          total++;
          state = 3;
          }
        break;
      } /* End switch */
    } /* End while */

  return total;
  }
 

